#include <cmath>

#include "Generator.hpp"
#include "Checker.hpp"
#include "Exceptions.hpp"

using SM = std::pair<Score, Move>;

Moves generate_for_pawn(const Board& board, const Coord& from) {
    const Color color = board.get_piece_at(from).color;
    const Color oponent_color = color.inverse();
    UNumber last_rank = (color == WHITE)? 7: 2;
    UNumber start_rank = (color == WHITE)? 2: 7;
    UNumber promotion_rank = (color == WHITE)? 8: 1;
    SNumber direction_rank = (color == WHITE)? 1: -1;

    auto base_score = Evaluation::piece(Piece(WHITE, PAWN));

    Moves result;

    // silence
    if (from.rank != last_rank) {
        if (!board.has_piece_at(Coord(from.file, from.rank+direction_rank))) {
            result.push_back(SM(base_score, Move::silent(from, Coord(from.file, from.rank+direction_rank))));
            if (from.rank == start_rank && !board.has_piece_at(Coord(from.file, from.rank+direction_rank*2)))
				result.push_back(SM(base_score + 1, Move::pawn_double(from, Coord(from.file, from.rank + direction_rank * 2))));
        }
    }

    // captures
    for (UNumber to_file = from.file - 1; to_file <= from.file + 1; to_file+=2 ) {
        if (to_file < 1 || to_file > 8 || from.rank == last_rank) continue;
        const Coord to = Coord(to_file, from.rank + direction_rank);
        if (board.has_piece_at(to) && board.get_piece_at(to).color == oponent_color) {
            auto victim_cost = Evaluation::piece(board.get_piece_at(to));
            result.push_back(SM(victim_cost * 10000 - base_score, Move::capture(from, to)));
        }

        // enpassant
        if (board.enpassant_posibility != EnpassantPosibility::none() && board.enpassant_posibility.coord.get() == to)
            result.push_back(SM(base_score * 10000 - base_score, Move::enpassant(from, to)));
    }

    // promotions
    if (from.rank == last_rank) {
        for (SNumber delta_file = -1; delta_file <= 1; ++delta_file) {
            if (from.file + delta_file > 8 || from.file + delta_file < 1) continue;
            const Coord to(from.file + delta_file, promotion_rank);

            if (delta_file == 0 && board.has_piece_at(to)) continue;
            else if (delta_file != 0 && (!board.has_piece_at(to) || board.get_piece_at(to).color != oponent_color)) continue;

            result.push_back(SM(Evaluation::piece(Piece(WHITE, KNIGHT)), Move::promotion(from, to, KNIGHT)));
            result.push_back(SM(Evaluation::piece(Piece(WHITE, BISHOP)), Move::promotion(from, to, BISHOP)));
            result.push_back(SM(Evaluation::piece(Piece(WHITE, ROOK)), Move::promotion(from, to, ROOK)));
            result.push_back(SM(Evaluation::piece(Piece(WHITE, QUEEN)), Move::promotion(from, to, QUEEN)));
        }
    }

    return result;
}

Moves generate_for_knight(const Board& board, const Coord& from) {
    Moves result;

    const Color color = board.get_piece_at(from).color;
    const Color oponent_color = color.inverse();

    auto base_score = Evaluation::piece(Piece(WHITE, KNIGHT));

    for (SNumber delta_file = -2; delta_file <= 2; ++delta_file) {
        for (SNumber delta_rank = -2; delta_rank <= 2; ++delta_rank) {
            if (std::abs(delta_file) + std::abs(delta_rank) != 3) continue;
            if (!Coord::is_valid(from.file + delta_file, from.rank + delta_rank)) continue;

            const Coord to(from.file + delta_file, from.rank + delta_rank);

            if (!board.has_piece_at(to)) {
                result.push_back(SM(base_score, Move::silent(from, to)));
            } else if (board.get_piece_at(to).color == oponent_color) {
                auto victim_cost = Evaluation::piece(board.get_piece_at(to));
                result.push_back(SM(victim_cost * 10000 - base_score, Move::capture(from, to)));
            }
        }
    }

    return result;
}

Moves generate_for_bishop(const Board& board, const Coord& from) {
    Moves result;

    const Color color = board.get_piece_at(from).color;
    const Color oponent_color = color.inverse();

    auto base_score = Evaluation::piece(Piece(WHITE, BISHOP));

    for (SNumber direction_file = -1; direction_file <= 1; ++direction_file) {
        for (SNumber direction_rank = -1; direction_rank <= 1; ++direction_rank) {
            if ((direction_file == 0) || (direction_rank == 0)) continue;

            for (SNumber delta = 1; delta <= 8; ++delta) {
                if (!Coord::is_valid(from.file+delta*direction_file, from.rank+delta*direction_rank)) break;
                auto to = Coord(from.file+delta*direction_file, from.rank+delta*direction_rank);

                if (!board.has_piece_at(to)) {
                    result.push_back(SM(base_score + delta, Move::silent(from, to)));
                } else {
                    if (board.get_piece_at(to).color == oponent_color) {
                        auto victim_cost = Evaluation::piece(board.get_piece_at(to));
                        result.push_back(SM(victim_cost * 10000 - base_score + delta, Move::capture(from, to)));
                    }
                    break;
                }
            }
        }
    }

    return result;
}

Moves generate_for_rook(const Board& board, const Coord& from) {
    Moves result;

    const Color color = board.get_piece_at(from).color;
    const Color oponent_color = color.inverse();

    auto base_score = Evaluation::piece(Piece(WHITE, ROOK));

    for (SNumber direction_file = -1; direction_file <= 1; ++direction_file) {
        for (SNumber direction_rank = -1; direction_rank <= 1; ++direction_rank) {
            if ((direction_file == 0) == (direction_rank == 0)) continue;

            for (SNumber delta = 1; delta <= 8; ++delta) {
                if (!Coord::is_valid(from.file+delta*direction_file, from.rank+delta*direction_rank)) break;
                auto to = Coord(from.file+delta*direction_file, from.rank+delta*direction_rank);

                if (!board.has_piece_at(to)) {
                    result.push_back(SM(base_score + delta, Move::silent(from, to)));
                } else {
                    if (board.get_piece_at(to).color == oponent_color) {
                        auto victim_cost = Evaluation::piece(board.get_piece_at(to));
                        result.push_back(SM(victim_cost * 10000 - base_score + delta, Move::capture(from, to)));
                    }
                    break;
                }
            }
        }
    }

    return result;
}

Moves generate_for_queen(const Board& board, const Coord& from) {
    Moves result;

    const Color color = board.get_piece_at(from).color;
    const Color oponent_color = color.inverse();

    auto base_score = Evaluation::piece(Piece(WHITE, QUEEN));

    for (SNumber direction_file = -1; direction_file <= 1; ++direction_file) {
        for (SNumber direction_rank = -1; direction_rank <= 1; ++direction_rank) {
            for (SNumber delta = 1; delta <= 8; ++delta) {
                if (!Coord::is_valid(from.file+delta*direction_file, from.rank+delta*direction_rank)) break;

                auto to = Coord(from.file+delta*direction_file, from.rank+delta*direction_rank);

                if (!board.has_piece_at(to)) {
                    result.push_back(SM(base_score + delta, Move::silent(from, to)));
                } else {
                    if (board.get_piece_at(to).color == oponent_color) {
                        auto victim_cost = Evaluation::piece(board.get_piece_at(to));
                        result.push_back(SM(victim_cost * 10000 - base_score + delta, Move::capture(from, to)));
                    }
                    break;
                }
            }
        }
    }

    return result;
}

Moves generate_for_king(const Board& board, const Coord& from) {
    Moves result;

    const Color color = board.get_piece_at(from).color;
    const Color oponent_color = color.inverse();

    auto base_score = Evaluation::piece(Piece(WHITE, QUEEN)) - 1;

    for (SNumber delta_file = -1; delta_file <= 1; ++delta_file) {
        for (SNumber delta_rank = -1; delta_rank <= 1; ++delta_rank) {
            if (!Coord::is_valid(from.file + delta_file, from.rank + delta_rank)) continue;

            const Coord to(from.file + delta_file, from.rank + delta_rank);

            if (!board.has_piece_at(to)) {
                result.push_back(SM(base_score, Move::silent(from, to)));
            } else if (board.get_piece_at(to).color == oponent_color) {
                auto victim_cost = Evaluation::piece(board.get_piece_at(to));
                result.push_back(SM(victim_cost * 10000 - base_score, Move::capture(from, to)));
            }
        }
	}
    // Rookings
	if (color == WHITE && board.rooking_posibility.white_king && from == Coord::from_fen("e1")) {
        auto to = Coord::from_fen("g1");
        auto rook_to = Coord::from_fen("f1");
		if (board.line_is_empty(from, to) && !board.has_piece_at(to)) {
			if (!Checker::is_controlled(board, to, oponent_color) &&
				!Checker::is_controlled(board, from, oponent_color) &&
				!Checker::is_controlled(board, rook_to, oponent_color)
				)
					result.push_back(SM(base_score - 1, Move::rooking(from, to)));
        }
	}

	if (color == WHITE && board.rooking_posibility.white_queen && from == Coord::from_fen("e1")) {
        auto to = Coord::from_fen("c1");
        auto rook_to = Coord::from_fen("d1");
		if (board.line_is_empty(from, to) && !board.has_piece_at(to) )
			if (!Checker::is_controlled(board, to, oponent_color) &&
				!Checker::is_controlled(board, from, oponent_color) &&
				!Checker::is_controlled(board, rook_to, oponent_color)
				)
					result.push_back(SM(base_score - 1, Move::silent(from, to)));
	}

	if (color == BLACK && board.rooking_posibility.black_king && from == Coord::from_fen("e8")) {
        auto to = Coord::from_fen("g8");
        auto rook_to = Coord::from_fen("f8");
		if (board.line_is_empty(from, to) && !board.has_piece_at(to))
			if (!Checker::is_controlled(board, to, oponent_color) &&
				!Checker::is_controlled(board, from, oponent_color) &&
				!Checker::is_controlled(board, rook_to, oponent_color)
				)
					result.push_back(SM(base_score - 1, Move::rooking(from, to)));
    }

	if (color == BLACK && board.rooking_posibility.black_queen && from == Coord::from_fen("e8")) {
        auto to = Coord::from_fen("c8");
        auto rook_to = Coord::from_fen("d8");
		if (board.line_is_empty(from, to) && !board.has_piece_at(to))
			if (!Checker::is_controlled(board, to, oponent_color) &&
				!Checker::is_controlled(board, from, oponent_color) &&
				!Checker::is_controlled(board, rook_to, oponent_color)
				)
					result.push_back(SM(base_score - 1, Move::rooking(from, to)));
	}

    return result;
}

bool inlegal(Board& board, const Move& move) {
	auto data = move.make(board);
	auto result = Checker::is_check(board, board.turn_color.inverse());
	data.restore(board);
    return result;
};

Moves Generator::for_piece(Board& board, const Coord& from) {
    Piece piece = board.get_piece_at(from);
    Moves result;

    if (piece.type == PAWN)        result = generate_for_pawn  (board, from);
    else if (piece.type == KNIGHT) result = generate_for_knight(board, from);
    else if (piece.type == BISHOP) result = generate_for_bishop(board, from);
    else if (piece.type == ROOK)   result = generate_for_rook  (board, from);
    else if (piece.type == QUEEN)  result = generate_for_queen (board, from);
	else if (piece.type == KING)   result = generate_for_king(board, from);
    else throw CustomException("Unknown piece type in generator. " + piece.to_fen());

    Moves filtered;
    for (auto scored_move: result) {
		auto move = scored_move.second;
        //if (Move::potentianal_inlegal(from, board))
		if (!inlegal(board, move)) filtered.push_back(scored_move);
	}

    return filtered;
}


Moves Generator::for_board(Board& board) {
	Moves result;

	for (UNumber file = 1; file <= 8; ++file) {
		for (UNumber rank = 1; rank <= 8; ++rank) {
			auto from = Coord(file, rank);
			if (!board.has_piece_at(from)) continue;
			auto piece = board.get_piece_at(from);
			if (piece.color == board.turn_color) {
				auto piece_moves = Generator::for_piece(board, from);
				result.splice(result.end(), piece_moves);
			}
		}
	}

	auto fstgrt = [](std::pair<Score, Move> a, std::pair<Score, Move> b) {return a.first > b.first;};

	result.sort(fstgrt);

    return result;
}
