#include <algorithm>
#include <iostream>
#include <vector>
#include <chrono>
#include <thread>
#include <cstdio>
#include <cmath>

#include "Search.hpp"
#include "Generator.hpp"
#include "Exceptions.hpp"
#include "Checker.hpp"

// for clang and gcc compilers
#define _GLIBCXX_USE_NANOSLEEP 1

using Search_Results = std::pair<Score, std::list<Move>>;

Search_Results quies(Board& board, Score alpha, Score beta, bool managed) {
	auto null = Search_Results(-Evaluation::max_score, std::list<Move>());

	if (managed) {
		++Search::Manager::nodes;
		if (Search::Manager::stoped) return null;
	}

	Score score = Evaluation::total(board);

	alpha = std::max(alpha, score);

	std::list<Move> pv;
    if (score >= beta) return Search_Results(score, pv);

	auto moves = Generator::for_board(board);
	for (auto scored_move : moves) {
		auto move = scored_move.second;
		if (!move.is_capture()) continue;

		auto restore_data = move.make(board);
		auto search_results = quies(board, -beta, -alpha, managed);
		score = -search_results.first;

		restore_data.restore(board);

		if (score > alpha) {
			alpha = score;
			pv = search_results.second;
			pv.push_front(move);
		}

        if (score >= beta)
            return Search_Results(score, pv);
	}

	return Search_Results(score, pv);
}

Search_Results negamax(Board& board, UNumber depth, bool managed) {
	auto null = Search_Results(-Evaluation::max_score, std::list<Move>());
	if (managed) {
        ++Search::Manager::nodes;
        if (Search::Manager::stoped) return null;
    }

    if (depth <= 0) return Search_Results(Evaluation::total(board), std::list<Move>());

    Score max_score = -Evaluation::max_score;
    std::list<Move> pv;

    auto moves = Generator::for_board(board);

    for (auto scored_move: moves) {
		auto move = scored_move.second;
		if (managed && Search::Manager::stoped) return null;
        auto restore_data = move.make(board);
        auto search_results = negamax(board, depth - 1, managed);
        auto score = -search_results.first;

        if (score > max_score) {
            max_score = score;
            pv = search_results.second;
            pv.push_front(move);
        }

        restore_data.restore(board);
    }

	if (moves.empty()) {
		if (Checker::is_check(board)) return Search_Results(-Evaluation::mate() + (Search::Manager::current_depth - depth), pv);
		return Search_Results(0, pv);
	}

    return Search_Results(max_score, pv);
}

Search_Results Search::negamax(Board& board, UNumber depth) {
	Search::Manager::current_depth = depth;
    return negamax(board, depth, false);
}

Search_Results alphabeta(Board& board, Score alpha, Score beta, UNumber depth, bool managed) {
    auto null = Search_Results(-Evaluation::max_score, std::list<Move>());
    if (managed) {
        ++Search::Manager::nodes;
        if (Search::Manager::stoped) return null;
    }

    if (depth <= 0) return quies(board, alpha, beta, managed);

    std::list<Move> pv;

    auto moves = Generator::for_board(board);

    for (auto scored_move: moves) {
		auto move = scored_move.second;
        if (managed && Search::Manager::stoped) return null;
        auto restore_data = move.make(board);
        auto search_results = alphabeta(board, -beta, -alpha, depth - 1, managed);
        auto score = -search_results.first;

        if (score > alpha) {
            alpha = score;
            pv = search_results.second;
            pv.push_front(move);
        }

        restore_data.restore(board);

        // alpha/beta pruning
        if (alpha >= beta) return Search_Results(alpha, std::list<Move>());
    }

	if (moves.empty()) {
		if (Checker::is_check(board)) return Search_Results(-Evaluation::mate() + (Search::Manager::current_depth - depth), pv);
		return Search_Results(0, pv);
	}

    return Search_Results(alpha, pv);
}

Search_Results alphabeta_root(Board& board, UNumber depth, bool managed) {
	Search::Manager::current_depth = depth;
	return alphabeta(board, -Evaluation::max_score, Evaluation::max_score, depth, managed);
}

Search_Results Search::alphabeta(Board& board, Score alpha, Score beta, UNumber depth) {
	Search::Manager::current_depth = depth;
    return alphabeta(board, alpha, beta, depth, false);
}

Search_Results negascout(Board& board, Score alpha, Score beta, UNumber depth, bool managed) {
    auto null = Search_Results(-Evaluation::max_score, std::list<Move>());
    if (managed) {
        ++Search::Manager::nodes;
        if (Search::Manager::stoped) return null;
    }

    if (depth <= 0) return quies(board, -beta, -alpha, managed);

    std::list<Move> pv;

    auto moves = Generator::for_board(board);

    bool first = true;

	for (auto scored_move : moves) {
		auto move = scored_move.second;
        if (managed && Search::Manager::stoped) return null;
        auto restore_data = move.make(board);
        auto search_results = first?
            negascout(board, -beta, -alpha, depth - 1, managed)
            :negascout(board, -(alpha+1), -alpha, depth - 1, managed);
        first = false;
        auto score = -search_results.first;

		if (score > alpha && score < beta) {
			search_results = negascout(board, -beta, -score, depth - 1, managed);
			score = -search_results.first;
		}

        if (score > alpha) {
            alpha = score;
            pv = search_results.second;
            pv.push_front(move);
        }

        restore_data.restore(board);

        // alpha/beta pruning
        if (alpha >= beta) return Search_Results(alpha, std::list<Move>());
    }

	if (moves.empty()) {
		if (Checker::is_check(board)) return Search_Results(-Evaluation::mate() + (Search::Manager::current_depth - depth), pv);
		return Search_Results(0, pv);
	}

    return Search_Results(alpha, pv);
}

Search_Results negascout_root(Board& board, UNumber depth, bool managed) {
	Search::Manager::current_depth = depth;
    return negascout(board, -Evaluation::max_score, Evaluation::max_score, depth, managed);
}

Search_Results Search::negascout(Board& board, Score alpha, Score beta, UNumber depth) {
	Search::Manager::current_depth = depth;
    return negascout(board, alpha, beta, depth, false);
}

Move Search::best_move(Board& board, UNumber depth, bool managed) {
    auto best_pair = negamax(board, depth, managed);
    if (best_pair.second.size() > 0)
        return best_pair.second.front();
    else
        return Move::none();
}

UNumber Search::Manager::depth = -1; // as infinity, not bug
UNumber Search::Manager::current_depth; // not initialized
UNumber Search::Manager::movetime = -1; // same
UNumber Search::Manager::nodes = 0;
Board Search::Manager::board = Board::empty();
bool Search::Manager::stoped = true;
Move Search::Manager::best_move = Move::none();

void Search::Manager::reset() {
    join_all();
	if (!stoped) throw CustomException("Can't run another search.");
	best_move = Move::none();
	depth = -1; // not error
    movetime = -1; // same
    nodes = 0;
	board = Board::empty();
}

std::vector<std::thread> works;

void stop_in(UNumber ms) {
    const UNumber accuracy = 100;

	while (ms > 0 && !Search::Manager::stoped) {
        std::chrono::milliseconds interval(accuracy);
        std::this_thread::sleep_for(interval);
        ms = (ms <= accuracy)? 0: ms - accuracy;
	}

    Search::Manager::stop();
    return;
}

void Search::Manager::start() {
	if (!stoped) throw CustomException("I can't search 2 best moves at time.");
	stoped = false;
	join_all();

    if (movetime != -1) works.push_back(std::thread(stop_in, movetime));

	works.push_back(std::thread(inc_search));
}

void Search::Manager::clear() {
	stop();
	for (auto& work : works) work.join();
	works.clear();
}

void Search::Manager::join_all() {
	for (auto& work : works) work.join();
	works.clear();
}

void Search::Manager::stop() {
	if (stoped) return;
	stoped = true;
	std::cout << "bestmove " << best_move.to_fen() << "\n" << std::flush;
}

void Search::Manager::inc_search() {
	for (UNumber d = 1; d <= depth; ++d) {
		current_depth = d;
        using namespace std::chrono;
        auto t1 = high_resolution_clock::now();
        auto search_results = alphabeta_root(board, d, true);
        auto t2 = high_resolution_clock::now();
        auto pv = search_results.second;
        best_move = (pv.size() > 0)? pv.front(): Move::none();
        auto score_i = search_results.first;
        std::string score_prefix = Evaluation::is_mate(score_i)? "mate": "cp";
        SNumber score_postfix = Evaluation::is_mate(score_i)? Evaluation::get_mate_half_moves(score_i): score_i / 10;
        double time = duration_cast<duration<double>>(t2 - t1).count();
        UNumber nps = Search::Manager::nodes / time;
        std::string pv_string;
        for (auto move: pv) pv_string += move.to_fen() + " ";

		if (stoped) return;
		std::cout << "info "
			<< "depth " << d << " "
			<< "score " << score_prefix << " " << score_postfix << " "
			<< "nps " << nps << " "
			<< "nodes " << nodes << " "
			<< "pv " << pv_string 
			<< "\n" << std::flush;

		// Todo: if debug mode
        //std::cout << "info string Branching Factor is " << std::pow(nodes, 1.0/d) << "\n";
	}

	stop();
}
