#include <iostream>
#include <vector>

#include <boost/algorithm/string.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/classification.hpp>
#include <chrono>

#include "UCI.hpp"
#include "Board.hpp"
#include "Exceptions.hpp"
#include "Move.hpp"
#include "Search.hpp"
#include "Time.hpp"
#include "Generator.hpp"

Board working_board;

const std::string Engine_Name = "ExCE";
const std::string Engine_Version = "1.1.8 (r118)";
const std::string Engine_Author = "Denis Gruzdev";

void id_answer() {
    std::cout << "id name " << Engine_Name << " " << Engine_Version << "\n";
    std::cout << "id author " << Engine_Author << "\n";
}

void option_answer() {
    // none
}

void uciok_answer() {
    std::cout << "uciok\n";
}

void unknown_answer() {
    std::cout << "Unknown command. Please try some else.\n";
}

void readyok_answer() {
	std::cout << "readyok\n";
}

void info_answer(std::string text) {
    std::cout << "info string " << text << "\n";
}

void debug_answer(std::string mode) {
    if (mode == "on")
        info_answer("Debug mode was enabled.");
    else if (mode == "off")
        info_answer("Debug mode was disabled.");
    else
        unknown_answer();
}

//todo: to Perft.{h|c}pp
UNumber perft(Board& board, UNumber depth) {
    if (depth == 0) return 1;

    auto moves = Generator::for_board(board);

    UNumber result = 0;
	for (auto scored_move : moves) {
		auto move = scored_move.second;
        auto data = move.make(board);
        auto value = perft(board, depth - 1);
        result += value;
        data.restore(board);
    }

    return result;
}

UNumber verbose_perft(Board& board, UNumber depth) {
    auto moves = Generator::for_board(board);

    UNumber result = 0;
    for (auto move_it = moves.begin(); move_it != moves.end(); ++move_it) {
        auto move = move_it->second;
        auto data = move.make(board);
        auto value = perft(board, depth - 1);
        std::cout << move.to_fen() << " - " << value << "\n";
        result += value;
        data.restore(board);
    }

    return result;
}

void Uci::perft_command(const std::vector<std::string>& parts) {
    if (parts.size() == 2) {
        using namespace std::chrono;
        auto depth = boost::lexical_cast<UNumber>(parts.at(1));
        auto t1 = high_resolution_clock::now();
        auto nodes = verbose_perft(working_board, depth);
        auto t2 = high_resolution_clock::now();
        double time = duration_cast<duration<double>>(t2 - t1).count();
        double nps = nodes / time;

        std::cout << "\n"
            << "    Total: " << nodes << " nodes\n"
            << "    Time: " << time << "sec\n"
            << "    Perfomance: " << nps / 1000 << "kN/sec\n";
    } else {
        unknown_answer();
    }
}

void Uci::debug_command(const std::vector<std::string>& parts) {
    if (parts.size() == 2) {
        debug_answer(parts.at(1));
    } else {
        unknown_answer();
    }
}

void Uci::go_command(const std::vector<std::string>& parts) {
	Search::Manager::reset();
    Search::Manager::board = working_board;
    Time::Manager::reset();
    Time::Manager::board = working_board;

	UNumber cursor = 0;
	for (UNumber i = cursor + 1; i < parts.size(); ++i) {
		if (parts.at(i) == "depth") {
			i++;
			const UNumber depth = boost::lexical_cast<UNumber>(parts.at(i));
			Search::Manager::depth = depth;
		}
        else if (parts.at(i) == "movetime"){
            i++;
            const UNumber movetime = boost::lexical_cast<UNumber>(parts.at(i));
            Search::Manager::movetime = movetime;
        }
		else if (parts.at(i) == "infinite"){
            // none, but works
		}
        else if (parts.at(i) == "wtime"){
            i++;
            const UNumber wtime = boost::lexical_cast<UNumber>(parts.at(i));
            Time::Manager::wtime = wtime;
        }
        else if (parts.at(i) == "btime"){
            i++;
            const UNumber btime = boost::lexical_cast<UNumber>(parts.at(i));
            Time::Manager::btime = btime;
        }
        else if (parts.at(i) == "winc"){
            i++;
            const UNumber winc = boost::lexical_cast<UNumber>(parts.at(i));
            Time::Manager::winc = winc;
        }
        else if (parts.at(i) == "binc"){
            i++;
            const UNumber binc = boost::lexical_cast<UNumber>(parts.at(i));
            Time::Manager::binc = binc;
        }
        else if (parts.at(i) == "movestogo"){
            i++;
            const UNumber movestogo = boost::lexical_cast<UNumber>(parts.at(i));
            Time::Manager::movestogo = movestogo;
        }
		else {
			unknown_answer();
			return;
		}
	}

    if (Search::Manager::movetime == -1)
        Search::Manager::movetime = Time::Manager::calculate();
	Search::Manager::start();
}

void Uci::ucinewgame_command() {
}

void Uci::isready_command() {
	readyok_answer();
}

void Uci::position_command(const std::vector<std::string>& parts) {
	if (parts.size() <= 1) {
		unknown_answer();
		return;
	}

	UNumber cursor = 0;

	if (parts.at(1) == "startpos") {
		working_board = Board::initial();
		cursor = 2;
	}
	else if (parts.at(1) == "fen") {
		std::string fen;
		for (UNumber i = 2; i < parts.size() && parts.at(i) != "moves"; ++i) {
			fen += parts.at(i) + " ";
			cursor = i + 1;
		}
		try {
			working_board = Board::from_fen(fen);
		}
		catch (...) {
			std::cout << "Bad fen.\n";
		}
	}

	if (parts.size() > cursor && parts.at(cursor) == "moves") {
		for (UNumber mi = cursor + 1; mi < parts.size(); ++mi) {
			auto move = Move::from_fen(parts.at(mi), working_board);
			move.make(working_board);
		}
	}
}

void Uci::uci_command() {
	id_answer();
	option_answer();
	uciok_answer();
}

bool process_command() {
	std::string command;
	std::getline(std::cin, command);

	std::vector<std::string> command_parts;
	boost::split(command_parts, command, boost::is_any_of(" \t\n\r"));
    if (command_parts.size() == 1 && command_parts.at(0).size() == 0) return true;

	if (command == "uci") Uci::uci_command();
	else if (command == "isready") Uci::isready_command();
	else if (command == "ucinewgame") Uci::ucinewgame_command();
	else if (command == "stop") Search::Manager::stop();
	else if (command == "ponderhit");
	else if (command == "quit") {
		Search::Manager::clear();
		return false;
	}

	else if (command_parts.at(0) == "go") Uci::go_command(command_parts);
	else if (command_parts.at(0) == "debug") Uci::debug_command(command_parts);
	else if (command_parts.at(0) == "position") Uci::position_command(command_parts);
	else if (command_parts.at(0) == "setoption");
	else if (command_parts.at(0) == "register");

	// custom commands
    else if (command_parts.at(0) == "perft") Uci::perft_command(command_parts);
	else if (command == "board") std::cout << working_board.to_fen() << "\n";
	else if (command == "pieces") {
		for (auto square: working_board.squares)
			std::cout << square.first.to_fen() << ", " << square.second.to_fen() << "\n";
	}
	else unknown_answer();

	return true;
}

void Uci::uci_cycle() {
	while (true) {
		try {
			bool next = process_command();
			if (!next) break;
		}
		catch (CustomException e) {
			std::cout << "There are some error while processing command: " << e.what() << "\n";
		}
	};
}
